#! /usr/local/bin/perl -w

# 1115, Tue 6 Jun 00 (PST), CAIDA
#
# Nevil Brownlee, ITSS Technology Development, The University of Auckland


# Counts samples with reqd distrib, for values of specified attribute
#
# Command line: 
#   ./plt_freq.pl  dns-root.dif xxx.plt  flowtype % toturnaroundtime1
#          from_file to_file  attrib +%/ distrib required
#
#   ./plt_freq.pl dns-root-n/sdsc/20001117.dif freq.txt flowclass=0..1 flowkind=1..13

use strict;
use Time::Local;

my ($from_file, $to_file, $c_interval);

$from_file = shift;  # First item from command line
if (defined($from_file)) {
   open(FROM, $from_file)
   or die "Can't open from_file: $!\n";
   }
else { printf "No from_file specified!\n"; }
 
$to_file = shift;  # Second item from command line
if (defined($to_file)) {
   open(TO, ">$to_file")
   or die "Can't open to_file: $!\n";
   }
else { printf "No to_file specified!\n"; }
print TO "##plt_freq  $from_file\n";

my $ntm_hdr = <FROM>;  # Read the input header record
if ($ntm_hdr =~ /-c(\d+) /) {
   $c_interval = $1;
   }

my $format;
for (;;) {  #  Find the #Format record
   $format = <FROM>;
   last if $format =~ /^#Format: /
   }
$format =~ s/^#Format: //;

my ($i, %format, %dist, $next);
my ($nf, @type, @dif);
# type: 0 = scalar, 1 = tcpdata, 2 = distribution
# dif:  1 = dif (to be summed), 0 = fixed

for ($nf = 0; $format =~ s/([\w\[\(]\w*[\]|\)]?)//; $nf++) {  # Split up fmt
   $next = $1;
   $type[$nf] = 0;
   if ($next =~ s/^\((.+)\)$/$1/ ) {
      $dist{$next} = $nf;
      $type[$nf] = 2;  # Distribution
      }
   elsif ($next =~ s/^\[(.+)\]$/$1/ ) {
      $type[$nf] = 1;  # TCPdata
      }
   $format{$next} = $nf;
   $dif[$nf] = 0;
   if ($next =~ /^d_/) {
      $dif[$nf] = 1;  # To be summed
      }
   }


my ($item, @attrb, @attst, @attvl, @att_str, @attvh);
my $errs = 0;

for (my $a = 0; $item = shift; ) {  # Grab items from command line
   my ($atr, $test, $val, $str_atr, $valh);

   last if $item eq '+';  # Stop when we find a '+'
   last if $item eq '%';  #    or a '%'
   last if $item eq '/';  #    or a '/'

   $item = lc($item);
   if ($item =~ s/\.\.(\d+)//) { $valh = $1; }
   else { $valh = ''; }
   if ($item =~ /^(\w+)([=!]+)([\w.]+)/) {
      $atr = $1;  $test = substr($2,0,1);  $val = $3;
      $str_atr = $val =~ /\./ ;
      if ($val > $valh) {
         print "Bad attribute range for $atr ($val > $valh)\n";
         ++$errs;
         }
      }
   else { $atr = $item;  $test = '=';  $val = ''; }

   if (defined $format{$atr} ) {
      $attrb[$a] = $atr;  $attst[$a] = $test;
      $attvl[$a] = $val;  $att_str[$a] = $str_atr;
      $attvh[$a] = $valh;
      printf "   %s %s %s %s\n", $atr, $test, $val,
         $valh ne '' ? ".. $valh" : '';
      printf TO "   %s %s %s %s\n", $atr, $test, $val,
         $valh ne '' ? ".. $valh" : '';
      ++$a;
      }
   else {
      print STDERR "unknown attrib '$atr'\n";
      ++$errs;
      }
   }

my @plot;

if (defined($item)) {
   for (my $p=0; $plot[$p] = shift; ++$p ) { # Grab distribs
      $plot[$p] = lc($plot[$p]);
      if (defined $format{$plot[$p]}) { 
         printf " %s", $plot[$p];
         }
      else { 
         print STDERR "Unknown distribution '$plot[$p]' <<<\n";
         ++$errs;
         }
      }
   }
pop @plot;   # Last shift will have returned NULL

if ($errs != 0) { exit; }

# %format       = index in records for attribtes (from #Format)
# @attrb,@attvl = array of attrib/value pairs for selecting flows
# $item         = +  Use distrib raw counts (i.e. 'as is')
#                 %  Convert each sample to pecentages
#                 /  Convert to percentages, but ignore oflo bucket
# @plot         = list of attribs to plot

#printf "attrib %s=%s, plot %s=%d\n",
#    $attrb[0], $attvl[0], $plot[0], $format{$plot[0]};

my $nbr = 0;

my $last_c_time = 0;  # Globals for get_time()
my $x;       # x value (time of day) for gnuplot
my $c_time;  # Collection time (seconds from 1970)

my %flows;  # Hash for flows in sample

my %samples;  # Nbr of samples each key appears in
my %counts;   # Total topdus for each key
my (%count_gta, %count_gtb);
my $ac = 15;  my $bc = 8;

my $topdux = $format{'d_topdus'};
# print "topdux=$topdux <<<\n";

my $first_sample = 1;

while (<FROM>) { # read  flows file

   chop;    # trim new line
   if (/^#/) { # starts with a #
      my $hrec = $_;
      if (s/^#Time: //) {
#         get_time();
	 } 
      elsif (s/^#EndData: //) {
	 if ($first_sample) {
	    $first_sample = 0;
	    }
	 else {
            foreach my $key (sort keys %flows) {
               my @f = @{ $flows{$key} };
               ++$samples{$key};
	       if ($f[$topdux] > $bc) { $count_gtb{$key} += $f[$topdux]; }
	       if ($f[$topdux] > $ac) { $count_gta{$key} += $f[$topdux]; }
               $counts{$key} += $f[$topdux];
               }
            }
         %flows = ();
         }

      next;
      }  # End of # record handling

   my $vref = get_flow();

   my $key = match_flow($vref);  # '' if we don't want the flow
   if ($key ne '') {
      my $aref = $flows{$key};
      if (!defined($aref)) {
         $flows{$key} = $vref;
         }
      else {
         print STDERR "multiple matching flows in sample, run at_merge.pl\n";
         }
      }

#   ++$nbr;
#last if ($nbr > 100) # { exit; }
   }

print TO "\n key";
for (my $j = 1; $j != scalar(@attrb); ++$j) { print TO '    '; }
print TO " samples  counts    > $bc   > $ac\n\n";

foreach my $key (sort keys %samples) {
   printf TO "%s  %6d  %6d %6d %6d\n", $key, $samples{$key},
      defined($counts{$key}) ? $counts{$key} : 0,
      defined($count_gtb{$key}) ? $count_gtb{$key} : 0,
      defined($count_gta{$key}) ? $count_gta{$key} : 0;
   }

# That's it
#--------------------------------------------------------------------------


sub match_flow{  # Returns flow key, '' if match fails
   my ($vref) = @_;

   my ($av, $match, $fail, $key);
   $fail = 0;  $key = '';
   for (my $q = 0; $q != scalar(@attrb); ++$q) {
      if ($attvl[$q] ne '') {
         $av = $vref->[$format{$attrb[$q]}];
         $key = sprintf "%s %3d", $key, $av;
         if ($attvh[$q] eq '') {
            $match = $att_str[$q] ? $av eq $attvl[$q] : $av == $attvl[$q];
            }
         else {
            $match = ($av >= $attvl[$q] && $av <= $attvh[$q]);
            }
         if ($match) {
            last if $fail = $attst[$q] eq '!';
            }
         else {
            last if $fail = $attst[$q] eq '=';
            }
         } 
      }

   return $fail ? '' : $key;
   }

sub get_flow{  # Take $_, return ref to array of values

   my @value = ();
   for ($i = 0;  s/([0123456789ABCDEF.-]+|\(|\[)//;  $i++) {
         # remove leading non-numerics followed by (, [ or a value
         # NOTE: .- appear in addresses, : appears in V6 addresses
      if ($1 eq '(') {  #  have distrib
	 s/\W*([^)]+)\)//;  # remove everything upto next ) save it in $1
	 my @v = split(/ +/, $1);   # put the distribution values in @v
	 my $v = {};                # $v points to a hash
	 if (($v->{Trans} = $v[0]) != 0 ) {  # construct structure
	    $v->{Scale} = $v[1];
	    $v->{Lower} = $v[2];
	    $v->{Upper} = $v[3];
	    $v->{Buckets} = $v[4];
	    $v->{P1} = $v[5];
	    $v->{P2} = $v[6];
	    $v->{P3} = $v[7];
	    $v->{Dist} = [splice(@v,8)]; # copies from element 8 to end
#            if ($v->{P1} != 0) {
#               $samples_per_int = $c_interval/$v->{P1};
#               $v->{Dist}->[0] %= $samples_per_int;
#                  # Ignore empty collection intervals
#               }
            $value[$i] = $v;
	    }
	 } 
      elsif ($1 eq '[') {  # TCPdata
	 s/([^\]]+)\]//;  # Remove everything upto next ], save it in $1 
         my @v = split(/ +/, $1);
	 $value[$i] = \@v;
         }
      else {
	 $value[$i] = $1;
	 }
      }
   return \@value;  # Return ref to array
   }
